;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                                                       ;;
;;;                					                  ;;
;;;                     University of Nairobi, Kenya                      ;;
;;;                       Copyright (c) 2004                              ;;
;;;                        All Rights Reserved.                           ;;
;;;                                                                       ;;
;;;  Permission is hereby granted, free of charge, to use and distribute  ;;
;;;  this software and its documentation without restriction, including   ;;
;;;  without limitation the rights to use, copy, modify, merge, publish,  ;;
;;;  distribute, sublicense, and/or sell copies of this work, and to      ;;
;;;  permit persons to whom this work is furnished to do so, subject to   ;;
;;;  the following conditions:                                            ;;
;;;   1. The code must retain the above copyright notice, this list of    ;;
;;;      conditions and the following disclaimer.                         ;;
;;;   2. Any modifications must be clearly marked as such.                ;;
;;;   3. Original authors' names are not deleted.                         ;;
;;;   4. The authors' names are not used to endorse or promote products   ;;
;;;      derived from this software without specific prior written        ;;
;;;      permission.                                                      ;;
;;;                                                                       ;;
;;;  THE UNIVERSITY OF NAIROBI AND THE CONTRIBUTOR OUTSIDE ECHO           ;;
;;;  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING      ;;
;;;  ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   ;;
;;;  SHALL THE UNIVERSITY OF NAIROBI  OR OUTSIDE ECHO       BE LIABLE     ;;
;;;  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES    ;;
;;;  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN   ;;
;;;  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,          ;;
;;;  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF       ;;
;;;  THIS SOFTWARE.                                                       ;;
;;;                                                                       ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;									  ;;
;;;  Authors: Mucemi Gakuru, Fredrick K. Iraki and Kamanda Ngugi	  ;;
;;;  Tokenization rules for Kiswahili Particularly numbers, abbreviations,;;
;;;  dates and currency. Also used as part of Kiswahili G2P tool.	  ;;
;;;  	 	  							  ;;
;;; 									  ;;
;;;									  ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                                ;;
;; 		UON_SWAHILI_TOKEN_TO_WORDS                        ;;
;;                                                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (UON_swahili_kw::token_to_words token name)

;; Return punctuations removed on chuncking to take care of rhythmical stress.
;; Then call swahili_token

   (if (string-matches (item.feat token "punc") "[\.\?\!\,\;\:]+")
	(set! name (string-append name (item.feat token "punc"))))
   (swahili_token token name))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                                ;;
;; 			SWAHILI_TOKEN                             ;;
;;                                                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

 (define (swahili_token token name)

  ;; SET DECIMAL TYPE
  ;;=================
   (set! deci "nanukta")

  (cond


  ;; SWAHILI ABBREVIATIONS
  ;;======================
    ((or  (string-matches name "Dr[\\.]?") 	     	
	  (string-matches name "Eng[\\.]?") 
	  (string-matches name "Mr[\\.]?") 
	  (string-matches name "n[\\.]?k[\\.]?") 
	  (string-matches name "Mrs[\\.]?") 
	  (string-matches name "Miss[\\.]?") 
          (string-matches name "St[\\.]?") 	     	
   	  (string-matches name "Prof[\\.]?")     	
   	  (string-matches name "Ksh[s]?[\\.]?")      	
   	  (string-matches name "Rand")     		
   	  (string-matches name "US$")      		
   	  (string-matches name "EUR"))

	  (append (cdr (assoc_string name swahili_abbrv))))


  ;; RHYTHMICAL STRESS 
  ;;==================
   ((or (string-matches name ".*[\\,\\;\\:\\.\\!\\?]+")  
        (string-matches name ".*[\\,\\;\\:\\.\\!\\?]+"))

	   (set! endm (car (reverse (symbolexplode name))))

        (if (string-matches name ".*\\(\\.[0-9]+\\)+.+")
        (begin
        	(set! wordp1 (string-after name "."))
        	(set! wordp2 (string-before name "."))
        	(set! wordp1 (string-before wordp1 endm))
       		(set! wordp (string-append wordp2 "."))
       		(set! wordp (string-append wordp wordp1)))
        (set! wordp (string-before name endm)))

        (if (not (string-matches wordp "[a-z]+"))
            (begin
        	    (set! name (car (reverse (swahili_token token wordp))))))
        	    (set! name (string-append name endm))
    (append
     	(reverse (cdr (reverse (swahili_token token wordp))))
	     (list (swahili_r_stress (downcase name)))))

  ;; SWAHILI PERCENT
  ;;================
  ((string-matches name ".*%$")
   (append
    (list "aslimia")
    (swahili_token token (string-before name "%"))))

  ;; SWAHILI TIME
  ;;=============
  ((string-matches name "[0-9]?[0-9][:\\.][0-9][0-9][AaPp][Mm]")  
   ;;  must be am/pm present for . to be acceptable separator
   (let (hours mins half sep (ttime (downcase name)))
     (if (string-matches ttime ".*:.*")
	 (set! sep ":")
	 (set! sep "."))
     (set! hours (string-append "tim" (remove_leading_zeros (string-before ttime sep))))

     (set! mins (string-after ttime sep))

     (if (string-matches ttime ".*am")
          (set! sep "am") (set! sep "pm"))
          (set! mins (string-before mins sep))

    (if (string-matches sep "am")
	(set! tx "asubuhi")(set! tx "alasiri"))
 
     (append
      (list "saa" )
      (cdr (assoc_string hours swahili_time))
      (list "na" "dakika")
      (swahili_number mins)
      (list "saa" "za")
      (list tx))))

  ((string-matches name "[0-9]?[0-9]:[0-9][0-9]")  ;; time
   (set! hours (string-append "tim" (remove_leading_zeros (string-before name ":"))))
   (append
     (list "saa")
     (cdr (assoc_string hours swahili_time))     (list "na" "dakika")
     (swahili_number (remove_leading_zeros (string-after name ":")))))

  ((string-matches name "[0-9][0-9]:[0-9][0-9]:[0-9][0-9]")  ;; exact time

   (set! hours (string-append "tim" (remove_leading_zeros (string-before name ":"))))
   (append
    (list "saa")
    (cdr (assoc_string hours swahili_time))
    (list "na" "dakika")
    (swahili_number (remove_leading_zeros (string-before (string-after name ":") ":")))
    (list "na" "sekunde")
    (swahili_number (remove_leading_zeros (string-after (string-after name ":") ":")))

   ))


  ;; SWAHILI DATES
  ;;==============

  ((string-matches name "[0-9]+s")  ;; e.g. 1950s
   (append
    (list "miaka" "ya")
    (swahili_number (string-before name "s"))))

  ((string-matches name "[0-9]+'s")  ;; e.g. 1950's
   (append
    (list "miaka" "ya")
    (swahili_number (string-before name "'s"))))


   ((string-matches name "[0-9][0-9]?/[0-9][0-9]?/[0-9][0-9]\\([0-9][0-9]\\)?")

     (set! num1 (string-before name "/"))
     (set! num2 (string-before (string-after name "/") "/"))
     (set! year (string-after (string-after name "/") "/"))
     (set! num3 (string-append "mon" (remove_leading_zeros num2)))
     (if (string-matches year "[0-9][0-9]")(set! year (string-append "20" year)))

     (append
       (list "tarehe")
       (swahili_number num1)
       (list "mwezi" "wa")
       (cdr (assoc_string num3 swahili_month))
       (list "mwaka" "wa")
       (swahili_number year)))

   ((string-matches name "[0-9][0-9]?\\-[0-9][0-9]?\\-[0-9][0-9]\\([0-9][0-9]\\)?")

     (set! num1 (string-before name "-"))
     (set! num2 (string-before (string-after name "-") "-"))
     (set! year (string-after (string-after name "-") "-"))
     (set! num3 (string-append "mon" (remove_leading_zeros num2)))
     (if (string-matches year "[0-9][0-9]")(set! year (string-append "20" year)))

     (append
       (list "tarehe")
       (swahili_number num1)
       (list "mwezi" "wa")
       (cdr (assoc_string num3 swahili_month))
       (list "mwaka" "wa")
       (swahili_number year)))

  ((string-matches name "[0-9]+-[0-9]+")
      (append
      (list "nambari")
      (swahili_number (string-before name "-"))
      (list "hadi" "nambari")
      (swahili_number (string-after name "-"))))


  ;; SWAHILI CURRENCY
  ;;=================
   ((string-matches name "[A-Z]*[\\$#Y£\\(USD\\)\\(KSH\\)\\(EURO\\)]\\([0-9]+\\,\\)?\\([0-9]+\\,\\)?[0-9]+\\(\\.[0-9]+\\)?")
    (set! deci "nasenti")
    (swahili_currency name))

  ;; SWAHILI FRACTION
  ;;=================
   ((string-matches name "[1-9][0-9]*/[1-9][0-9]*")
    (swahili_fraction name))

  ;; SWAHILI NUMBER
  ;;===============
   ((string-matches name "\\([0-9]+\\,\\)?\\([0-9]+\\,\\)?[0-9]+\\(\\.[0-9]+\\)?")
	   (assign_number name))

  ;; ACCRONYMS
  ;;===============
   ((string-matches name "[B-DF-HJ-NP-TV-Zb-df-hj-np-tv-z]+") 
    (symbolexplode (downcase name)))

  ;; SWAHILI WORD
  ;;===============
   ((string-matches (downcase name) "[A-Za-z\\']+") 
    (take_careof_ng name)
    (list (downcase name)))


  ;; UNRESOLVED TOKEN TAKEN AS SYMBOLS
  ;;==================================
   (t (symbolexplode (downcase name))) 

))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                                ;;
;; 		  ASSIGN_SWAHILI_NUMBER                           ;;
;;                                                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(define (assign_number name)
    (if (string-matches name "\\([0-9]+\\,\\)?\\([0-9]+\\,\\)?[0-9]+")
            (set! name (string-append name ".00")))
    (cond
        ((string-matches name "\\([0-9]+\\,\\)+\\([0-9]+\\,\\)+[0-9]+\\(\\.[0-9]+\\)+")
     		(set! nmx1 (string-after  name ","))
     		(set! nmx2 (string-after  nmx1 ","))
     		(set! nmx3 (string-after  nmx2 ".")))

        ((string-matches name "\\([0-9]+\\,\\)+[0-9]+\\(\\.[0-9]+\\)+")
		(set! nmx1 "")
     		(set! nmx2 (string-after  name ","))
     		(set! nmx3 (string-after  nmx2 ".")))


        ((string-matches name "[0-9]+\\(\\.[0-9]+\\)+")
    		(set! nmx1 "")
    		(set! nmx2 name)
    		(set! nmx3 (string-after  nmx2 "."))))

    (set! nmx4 (string-before name ","))
    (set! nmx5 (string-before nmx1 ","))
    (set! nmx6 (string-before nmx2 "."))
    (set! nmx7 (string-append  nmx4 nmx5))
    (set! nmx8 (string-append  nmx7 nmx6))
    (append
    (swahili_number nmx8)
    (if (string-equal nmx3 "00") nil (append (list deci) (swahili_number nmx3)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                                ;;
;; 			SWAHILI_NUMBER                            ;;
;;                                                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  
(define (swahili_number name) "Convert a string of digits into a list of words saying the number."
  (if (string-matches name "0")
      (list "sufuri")
      (swahili_number_from_digits (symbolexplode name))))

(define (just_zeros digits) "If this only contains 0s then we just do something different."
 (cond
  ((not digits) t)
   ((string-equal "0" (car digits))
    (just_zeros (cdr digits)))
    (t nil)))

(define (swahili_number_from_digits digits)
  "Takes a list of digits and converts it to a list of words saying the number."
  (let ((l (length digits)))
    (cond

     ((equal? l 0)nil)

     ((string-equal (car digits) "0")
      (swahili_number_from_digits (cdr digits)))

     ((equal? l 1);; single digit
      (cond 
       ((string-equal (car digits) "0") (list "sufuri"))
       ((string-equal (car digits) "1") (list "moja"))
       ((string-equal (car digits) "2") (list "mbili"))
       ((string-equal (car digits) "3") (list "tatu"))
       ((string-equal (car digits) "4") (list "nne"))
       ((string-equal (car digits) "5") (list "tano"))
       ((string-equal (car digits) "6") (list "sita"))
       ((string-equal (car digits) "7") (list "saba"))
       ((string-equal (car digits) "8") (list "nane"))
       ((string-equal (car digits) "9") (list "tisa"))
       ))

     ((equal? l 2);; less than 100
      (cond
       ((string-equal (car digits) "0");; 0x
	(swahili_number_from_digits (cdr digits)))
     
        ((string-equal (car digits) "1");; 1x
	(if (string-equal (car (cdr digits)) "0") 
	    (list "kumi")
	    (append
	     (list "kumi")
             (list "na")
	     (swahili_number_from_digits (cdr digits)))))
    
       ((string-equal (car digits) "2");; 2x
	(if (string-equal (car (cdr digits)) "0") 
	    (list "ishirini")
	    (append
	     (list "ishirini")
             (list "na")
	     (swahili_number_from_digits (cdr digits)))))

       ((string-equal (car digits) "3");; 3x
	(if (string-equal (car (cdr digits)) "0") 
	    (list "therathini")
	    (append
	     (list "therathini")
             (list "na")
	     (swahili_number_from_digits (cdr digits)))))

       ((string-equal (car digits) "4");; 4x
	(if (string-equal (car (cdr digits)) "0") 
	    (list "arubaini")
	    (append
	     (list "arubaini")
             (list "na")
	     (swahili_number_from_digits (cdr digits)))))

       ((string-equal (car digits) "5");; 5x
	(if (string-equal (car (cdr digits)) "0") 
	    (list "hamsini")
	    (append
	     (list "hamsini")
             (list "na")
	     (swahili_number_from_digits (cdr digits)))))

       ((string-equal (car digits) "6");; 6x
	(if (string-equal (car (cdr digits)) "0") 
	    (list "sitini")
	    (append
	     (list "sitini")
             (list "na")
	     (swahili_number_from_digits (cdr digits)))))

       ((string-equal (car digits) "7");; 7x
	(if (string-equal (car (cdr digits)) "0") 
	    (list "sabini")
	    (append
	     (list "sabini")
             (list "na")
	     (swahili_number_from_digits (cdr digits)))))

       ((string-equal (car digits) "8");; 8x
	(if (string-equal (car (cdr digits)) "0") 
	    (list "themanini")
	    (append
	     (list "themanini")
             (list "na")
	     (swahili_number_from_digits (cdr digits)))))

       ((string-equal (car digits) "9");; 9x
	(if (string-equal (car (cdr digits)) "0") 
	    (list "tisaini")
	    (append
	     (list "tisaini")
             (list "na")
	     (swahili_number_from_digits (cdr digits)))))

       ))


    ((equal? l 3);; in the hundreds
      (let (	(sub_hundreds 
	     		(list 
	      		  (car (cdr (reverse digits)))
	      		  (car (reverse digits))
              		))
	    	(hundreds (reverse (cdr (cdr (reverse digits)))))
             )

	(append
	  (list "mia")
	  (swahili_number_from_digits hundreds)
          (if (just_zeros sub_hundreds) nil (list "na"))
 	  (swahili_number_from_digits sub_hundreds)
	)))
 
     ((< l 6)
      (let (	(sub_thousands 
	     		(list 
	      		  (car (cdr (cdr (reverse digits))))
	      		  (car (cdr (reverse digits)))
	      		  (car (reverse digits))
              		))
	    	(thousands (reverse (cdr (cdr (cdr (reverse digits))))))
             )

	(append
	  (list "elfu")
	  (swahili_number_from_digits thousands)
          (if (just_zeros sub_thousands) nil (list "na"))
 	  (swahili_number_from_digits sub_thousands)
	)))

     ((< l 7)
      (let (	(sub_laki 
	     		(list 
	                  (car (cdr (cdr (cdr (cdr (reverse digits))))))
	                  (car (cdr (cdr (cdr (reverse digits)))))
	      		  (car (cdr (cdr (reverse digits))))
	      		  (car (cdr (reverse digits)))
	      		  (car (reverse digits))
              		))
	    (laki (reverse (cdr (cdr (cdr (cdr (cdr (reverse digits)))))))))

	(append
	  (list "laki")
	  (swahili_number_from_digits laki)
          (if (just_zeros sub_laki) nil (list "na"))
 	  (swahili_number_from_digits sub_laki)
	)))

     
     ((< l 13)
      (let ((sub_million 
	     (list 
	      (car (cdr (cdr (cdr (cdr (cdr (reverse digits)))))))
	      (car (cdr (cdr (cdr (cdr (reverse digits))))))
	      (car (cdr (cdr (cdr (reverse digits)))))
	      (car (cdr (cdr (reverse digits))))
	      (car (cdr (reverse digits)))
	      (car (reverse digits))
	      ))
	    (millions (reverse (cdr (cdr (cdr (cdr (cdr (cdr (reverse digits))))))))))
	(set! x (swahili_number_from_digits millions))

	(append
	     (append (list "milioni") x)
	     (if (just_zeros sub_million) nil (list "na"))
	     (swahili_number_from_digits sub_million))))

     )))

      
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                                ;;
;; 			SWAHILI_CURRENCY                          ;;
;;                                                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (swahili_currency name)
"Returns a list of words for NAME from currency.  This allows the
user to customize various non-local, multi-word, context dependent
translations of tokens into words.  If this function is unset only
the builtin translation rules are used, if this is set the builtin
rules are not used unless explicitly called. [see Token to word rules]"

(set! currency_type 
      '(("US"  "marekani")
	("CAN" "canada")
	("JP"  "japa2ni")
	("SA"  "afri2kakusini")
	("A"   "ausitira2lia")
	("AUS" "ausitira2lia")
	("ZIM" "zimba2bwe") 
	("K"   "ke2nya")
	("U"   "uga2nda")
	("T"   "tanzani2a")
	("GB"  "uingere2za")
	("STG" "uingere2za")
	("UK"  "uingere2za")))

   (let (amount type currency)
     (cond
      ((string-matches name ".*\\$.*")
       (set! amount (string-after name "$"))
       (set! type (string-before name "$"))
       (set! currency "do2la"))
      ((string-matches name ".*£.*")
       (set! amount (string-after name "£"))
       (set! deci "napeni")
       (set! type (string-before name "£"))
       (set! currency "pau2ni"))
      ((string-matches name ".*#.*")
       (set! amount (string-after name "#"))
       (set! type (string-before name "#"))
       (set! deci "napeni")
       (set! currency "pau2ni"))
      ((string-matches name ".*Y.*")
       (set! amount (string-after name "Y"))
       (set! type "JP")
       (set! currency "ye2ni"))
      ((string-matches name ".*\\\\.*")
       (set! amount (string-after name "\\"))
       (set! type (string-before name "\\"))
       (set! currency "ye2ni"))
      ((string-matches name ".*\\(GBP\\).*")
       (set! amount (string-after name "GBP"))
       (set! type (string-before name "P"))
       (set! deci "napeni")
       (set! currency "pauni"))
      ((string-matches name ".*\\(USD\\).*")
       (set! amount (string-after name "USD"))
       (set! type (string-before name "D"))
       (set! currency "dola"))
      ((string-matches name ".*\\(EUR\\).*")
       (set! amount (string-after name "EUR"))
       (set! type (string-before name "EUR"))
       (set! currency "yuro"))
      ((string-matches name ".*\\(R\\).*")
       (set! amount (string-after name "R"))
       (set! type "SA")
       (set! currency "randi"))
      ((string-matches name ".*\\(SH\\).*[0-9].*")
       (set! amount (string-after name "SH"))
       (set! type (string-before name "SH"))
       (set! currency "shili2ngi"))
      (t
       ;; who knows
       (set! amount (string-after name "$"))
       (set! type (string-before name "$"))
       (set! currency "dola")))
       
     (if (string-matches type "[A-Za-z]+")  
  
          (append
 	    
            (list (string-append currency "za"))
            (cdr(assoc_string type currency_type))
            (assign_number amount))
          
          (append
            (list currency)
            (assign_number amount))
     )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                                ;;
;; 			SWAHILI_FRACTION                          ;;
;;                                                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (swahili_fraction name)

   (let ((numerator (string-before name "/"))
	 (denominator (string-after name "/")))

     (cond
      ((string-matches name "1/2")
       (list "nusu"))

      ((string-matches denominator "4")
       (append
	(list "robo")
	(if (string-equal numerator "1")
             nil (swahili_number numerator))))
      (t
       (append
	(swahili_number numerator)
	(list "kwa")
	(swahili_number denominator)))
)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                                ;;
;; 			SWAHILI_ABBREVIATIONS                     ;;
;;                                                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

   (set! swahili_abbrv 

	     '(("US"  	"marekaa2ni")
		("CAN" 	"kanada")
		("JP"  	"japani")
		("SA"  	"afri2kakusini")
		("AUS" 	"ausitiralia")
		("ZIM" 	"zimbabwe") 
		("K"   	"kenya")
		("U"   	"uganda")
		("T"   	"tanzania")
		("GB"  	"uingereza")
		("STG" 	"uingereza")
		("UK"  	"uingereza")
                ("Bw."	"bwana")
                ("Bw"	"bwana")
	   	("Dr."	"daktari")
	   	("Dr"   "daktari")
                ("nk"   "na" "kadha2li2ka")
                ("n.k."   "na" "kadha2li2ka")
                ("n.k"   "na" "kadha2li2ka")
                ("nk."   "na" "kadha2li2ka")
	   	("Mr."	"bwana")
	   	("Mr"   "bwana")
	   	("Mrs."	"bibi")
	   	("Mrs"  "bibi")
	   	("Miss"	"binti")
                ("MISS"	"binti")
	   	("Eng"  "mhandisi")
	   	("Eng."  "mhandisi")
	   	("St"	"mutakatifu")
	   	("St."	"mutakatifu")
	   	("Prof"	"profesa")
	   	("Ksh"	"shili2ngiza" "ke2nya")
	   	("Ksh."	"shili2ngiza" "ke2nya")
	   	("Kshs"	"shili2ngiza" "ke2nya")
	   	("Kshs." "shili2ngiza" "ke2nya")
		("Rand"	"randiza" "afrikakusini")
	   	("US$" "do2laza" "mare2kani")
	   	("EUR" 	"yu2ro")))


   (set! swahili_month

	      '(("mon1"  	"januari")
		("mon2"  	"februari")
		("mon3"  	"machi")
		("mon4"  	"aprili")
		("mon5"  	"mei")
		("mon6"  	"juni")
		("mon7"  	"julai")
		("mon8"  	"agosti")
		("mon9"  	"septemba")
		("mon10"  	"oktoba")
		("mon11"  	"novemba")
		("mon12"  	"desemba")))

   (set! swahili_time

	      '(("tim1"  	"saba")
		("tim2"  	"nane")
		("tim3"  	"tisa")
		("tim4"  	"kumi")
		("tim5"  	"kumi "na" "moja")
		("tim6"  	"kumi "na" "mbili")
		("tim7"  	"moja")
		("tim8"  	"mbili")
		("tim9"  	"tatu")
		("tim10"  	"nne")
		("tim11"  	"tano")
		("tim12"  	"sita")
	        ("tim13"  	"saba")
		("tim14"  	"nane")
		("tim15"  	"tisa")
		("tim16"  	"kumi")
		("tim17"  	"kumi"  "na" "moja")
		("tim18"  	"kumi" "na" "mbili")
		("tim19"  	"moja")
		("tim20"  	"mbili")
		("tim21"  	"tatu")
		("tim22"  	"nne")
		("tim23"  	"tano")
		("tim24"  	"sita")))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                                ;;
;; 			TAKE_CARE_OF_ng'                          ;;
;;                                                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (take_careof_ng name)
   (set! name2 name)
   (set! prex2 "")
   (cond 
	((string-matches (downcase name) ".*ng\\'.*") 
               	    (set! prex1 (string-before name "ng'"))
        	    (set! name1 (string-append prex1 "ng'"))
        	    (set! prex2 (string-after name "ng'"))
        	    (set! name2 (string-append prex1 "ng1"))
        	    (set! name (string-append name2 prex2))
		    (take_careof_ng name)))
        	    (set! name (string-append name2 prex2)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                                ;;
;; 			RHYTHMICAL_STRESS                         ;;
;;                                                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (swahili_r_stress name)
     (set! endm (car (reverse (symbolexplode name))))
     (set! wordx (string-before name endm))
     (set! endm1 (car (reverse (symbolexplode wordx))))
     (set! endmp (string-append endm1 endm))

   
     (if (and (> (length wordx) 1) (string-matches endm1 "[aeiou]+"))
        (begin
         (set! word1 (string-before name endmp))
         (set! endm2 (car (reverse (symbolexplode word1))))
         (set! endmp (string-append endm2 endmp))
 
     	 (if (string-matches endm2 "[aeiou]+") 
             (set! wordx (string-append word1 "2" endm1))

             (begin

     	     (if (> (length wordx) 2)
                (begin
         	 (set! word1 (string-before name endmp))
                 (set! endm3 (car (reverse (symbolexplode word1))))
         	 (set! endmp (string-append endm3 endmp))

   		 (cond
     			((string-matches endm3 "[aeiou]+") 
			 (set! wordx (string-append word1 "2" endm2  endm1)))

     			((and 
			 (string-matches endm3 "[n]+") (not (string-matches endm2 "[gyw]+")))
			 (set! wordx (string-append word1 "2" endm2  endm1)))

     			((and 
			 (string-matches endm3 "[m]+") (not (string-matches endm2 "[w]+")))
			 (set! wordx (string-append word1 "2" endm2  endm1)))
                        
                        (t 
      	         
			 (if (> (length wordx) 3)
                           (begin
         	            (set! word1 (string-before name endmp))
                            (set! endm4 (car (reverse (symbolexplode word1))))
         	            (set! endmp (string-append endm4 endmp))

   		            (cond
     				((string-matches endm4 "[aeiou]+") 
			 	 (set! wordx (string-append word1 "2" endm3 endm2  endm1)))

     				((and 
			 	 (string-matches endm4 "[n]+") 
				 (not (string-matches wordx ".*[gy]+[w\\']+[aeiou]+")))
			 	 (set! wordx (string-append word1 "2" endm3 endm2  endm1)))

     				((string-matches endm4 "[m]+") 
			 	 (set! wordx (string-append word1 "2" endm3 endm2  endm1)))

                        	(t 
      	         
			 	 (if (> (length wordx) 4)
                           	  (begin
         	            	    (set! word1 (string-before name endmp))
                            	    (set! endm5 (car (reverse (symbolexplode word1))))
         	            	    (set! endmp (string-append endm4 endmp))

   		            	    (cond
     					((string-matches endm5 "[aeiou]+") 
			 	 	(set! wordx 
					(string-append word1 "2" endm4 endm3 endm2  endm1)))

     					((string-matches endm5 "[m]+") 
			 	 	(set! wordx 
					(string-append word1 "2" endm4 endm3 endm2  endm1)))

     					((and 
			 	 	 (string-matches endm5 "[n]+") 
				 	 (not (string-matches wordx ".*gw\\'[aeiou]+")))
			 	 	 (set! wordx 
					 (string-append word1 "2" endm4 endm3 endm2  endm1)))
                                     ))                 	   
                   	             (set! wordx (string-append word1 endm3 endm2 endm1 "2"))
			  ))))                 	   
                   	   (set! wordx (string-append word1 endm2 endm1 "2"))
			  ))))

	         (set! wordx (string-append word1 endm1 "2"))
	     )))))
	     (set! name wordx)
	     (take_careof_ng name))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                                ;;
;; 		SELECT, RESET AND PROVIDE TOKENIZER               ;;
;;                                                                ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (UON_swahili_kw::select_tokenizer)
  "(UON_swahili_kw::select_tokenizer) Set up tokenizer for swahili."
  (Parameter.set 'Language 'UON_swahili)
  (set! token_to_words UON_swahili_kw::token_to_words))

(define (UON_swahili_kw::reset_tokenizer)
  "(UON_swahili_kw::reset_tokenizer) Reset any globals modified for this voice.  Called by 
   (UON_swahili_kw::voice_reset)."
  ;; None

  t)

(provide 'UON_swahili_kw_tokenizer)
